@@ -52,16 +52,16 @@
 #include <strings.h>
 
 #ifdef __MACH__
+#include <mach-o/dyld.h>
 #include <mach-o/fat.h>
 #include <mach-o/loader.h>
 
 #include <libkern/OSByteOrder.h>
-
-#include <dlfcn.h>
 #endif
 
 #include "libmachista.h"
 #include "hashmap.h"
+#include "strlcpy.h"
 
 #ifdef __MACH__
 /* Tiger compatibility */
@@ -467,11 +467,15 @@ int macho_parse_file(macho_handle_t *handle, const char *filepath, const macho_t
     
     /* Open input file */
     if ((fd = open(filepath, O_RDONLY)) < 0) {
-#ifdef HAVE_DLOPEN_PREFLIGHT
-        if (dlopen_preflight(filepath)) {
-            return MACHO_ECACHE;
+#ifdef HAVE__DYLD_SHARED_CACHE_CONTAINS_PATH
+    /* All systems that have this function in the SDK should also have
+       __builtin_available, so not bothering to check for it. */
+        if (__builtin_available(macos 11.0, ios 14.0, watchos 7.0, tvos 14.0, *)) {
+            if (_dyld_shared_cache_contains_path(filepath)) {
+                return MACHO_ECACHE;
+            }
         }
-#endif /* HAVE_DLOPEN_PREFLIGHT */
+#endif /* HAVE__DYLD_SHARED_CACHE_CONTAINS_PATH */
         return MACHO_EFILE;
     }
 
@@ -506,7 +510,7 @@ int macho_parse_file(macho_handle_t *handle, const char *filepath, const macho_t
             ret = MACHO_EMEM;
         }
     } else {
-        /* An error occured, free mt */
+        /* An error occurred, free mt */
         free_macho_t((macho_t *)*res);
         *res = NULL;
     }
